package ch.psi.daq.queryrest.controller.validator;

import java.util.ArrayList;
import java.util.LinkedHashSet;
import java.util.Set;

import org.springframework.beans.BeansException;
import org.springframework.context.ApplicationContext;
import org.springframework.context.ApplicationContextAware;
import org.springframework.validation.Errors;
import org.springframework.validation.Validator;

import ch.psi.daq.domain.backend.Backend;
import ch.psi.daq.domain.config.DomainConfig;
import ch.psi.daq.domain.query.DAQQueries;
import ch.psi.daq.domain.query.DAQQuery;
import ch.psi.daq.domain.query.DAQQueryElement;
import ch.psi.daq.domain.query.operation.Aggregation;
import ch.psi.daq.domain.query.operation.EventField;
import ch.psi.daq.domain.query.transform.ExecutionEnvironment;
import ch.psi.daq.domain.query.transform.ValueTransformationSequence;
import ch.psi.daq.domain.request.Request;
import ch.psi.daq.queryrest.config.QueryRestConfig;

public class EventQueryValidator implements Validator, ApplicationContextAware {
   private Set<EventField> defaultResponseFields;
   private Set<Aggregation> defaultResponseAggregations;

   @SuppressWarnings("unchecked")
   @Override
   public void setApplicationContext(ApplicationContext context) throws BeansException {
      final Backend backend = context.getBean(DomainConfig.BEAN_NAME_BACKEND_DEFAULT, Backend.class);
      context = backend.getApplicationContext();

      defaultResponseFields = context.getBean(QueryRestConfig.BEAN_NAME_DEFAULT_EVENT_RESPONSE_FIELDS, Set.class);
      defaultResponseAggregations = context.getBean(QueryRestConfig.BEAN_NAME_DEFAULT_EVENT_RESPONSE_AGGREGATIONS, Set.class);
   }

   @Override
   public boolean supports(final Class<?> clazz) {
      return DAQQuery.class.isAssignableFrom(clazz) || DAQQueries.class.isAssignableFrom(clazz);
   }

   @Override
   public void validate(final Object target, final Errors errors) {
      if (target instanceof DAQQuery) {
         this.checkElement((DAQQuery) target, errors);
      } else if (target instanceof DAQQueries) {
         final DAQQueries queries = (DAQQueries) target;
         for (final DAQQueryElement daqQueryElement : queries) {
            this.checkElement(daqQueryElement, errors);
         }
      }
   }

   private void checkElement(final DAQQueryElement query, final Errors errors) {
      // set default values (if not set)
      if (query.getFields() == null || query.getFields().isEmpty()) {
         query.setFields(new LinkedHashSet<>(defaultResponseFields));
      }

      if (query.getAggregation() != null) {
         // check if only one binning element is defined
         final long durationPerBin = query.getAggregation().getDurationPerBin();
         final long pulsesPerBin = query.getAggregation().getPulsesPerBin();
         final int nrOfBins = query.getAggregation().getNrOfBins();
         if ((durationPerBin != Request.NOT_SET && (pulsesPerBin != Request.NOT_SET || nrOfBins != Request.NOT_SET))
               || (pulsesPerBin != Request.NOT_SET
                     && (durationPerBin != Request.NOT_SET || nrOfBins != Request.NOT_SET))
               || (nrOfBins != Request.NOT_SET
                     && (durationPerBin != Request.NOT_SET || pulsesPerBin != Request.NOT_SET))) {
            errors.reject("durationPerBin", "Only one binning element must be defined.");
            errors.reject("pulsesPerBin", "Only one binning element must be defined.");
            errors.reject("nrOfBins", "Only one binning element must be defined.");
         }

         if (query.getRange().isPulseIdRangeDefined() && durationPerBin != Request.NOT_SET) {
            errors.reject("durationPerBin", "Pulse range queries only support pulse based binning.");
         }
         if (query.getRange().isTimeRangeDefined() && pulsesPerBin != Request.NOT_SET) {
            errors.reject("pulsesPerBin", "Time range queries only support time based binning.");
         }


         // set default values (if not set)
         if (query.getAggregation().getAggregations() == null || query.getAggregation().getAggregations().isEmpty()) {
            query.getAggregation().setAggregations(new ArrayList<>(defaultResponseAggregations));
         }

         // without adding this field, user need to explicitly ask for value field when querying
         // aggregations.
         // if (!query.getFields().contains(QueryField.value)) {
         // // without this field, json will not contain Stats
         // query.addField(QueryField.value);
         // }
      }

      if (query.getValueTransformations() != null && !query.getValueTransformations().isEmpty()) {
         // without this field, json will not contain transformedValue
         query.addField(EventField.transformedValue);

         for (final ValueTransformationSequence transformationSequence : query.getValueTransformations()) {
            transformationSequence.setExecutionEnvironment(ExecutionEnvironment.QUERYING);
         }
      }
   }
}
