package ch.psi.daq.queryrest.response.formatter;

import java.io.OutputStream;
import java.util.Iterator;
import java.util.LinkedHashMap;
import java.util.LinkedHashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.Map.Entry;
import java.util.concurrent.atomic.AtomicReference;
import java.util.function.Function;
import java.util.function.ToLongFunction;
import java.util.stream.Collectors;
import java.util.stream.Stream;

import org.apache.commons.lang3.tuple.Triple;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.beans.BeansException;
import org.springframework.context.ApplicationContext;
import org.springframework.context.ApplicationContextAware;

import com.fasterxml.jackson.core.JsonEncoding;
import com.fasterxml.jackson.core.JsonFactory;
import com.fasterxml.jackson.core.JsonGenerator;
import com.fasterxml.jackson.databind.ObjectMapper;
import com.fasterxml.jackson.databind.ObjectWriter;
import com.fasterxml.jackson.databind.ser.impl.SimpleBeanPropertyFilter;
import com.fasterxml.jackson.databind.ser.impl.SimpleFilterProvider;

import ch.psi.daq.common.stream.match.ListCreator;
import ch.psi.daq.common.stream.match.ListFiller;
import ch.psi.daq.common.stream.match.Padder;
import ch.psi.daq.common.stream.match.StreamMatcher;
import ch.psi.daq.common.time.TimeUtils;
import ch.psi.daq.domain.DataEvent;
import ch.psi.daq.domain.json.ChannelName;
import ch.psi.daq.domain.query.DAQQueryElement;
import ch.psi.daq.domain.query.backend.BackendQuery;
import ch.psi.daq.domain.query.bin.BinningStrategy;
import ch.psi.daq.domain.query.bin.strategy.BinningStrategyPerBinPulse;
import ch.psi.daq.domain.query.bin.strategy.BinningStrategyPerBinTime;
import ch.psi.daq.domain.query.mapping.IncompleteStrategy;
import ch.psi.daq.domain.query.mapping.Mapping;
import ch.psi.daq.domain.query.operation.Aggregation;
import ch.psi.daq.domain.query.operation.EventField;
import ch.psi.daq.domain.query.operation.Extrema;
import ch.psi.daq.domain.query.operation.QueryField;
import ch.psi.daq.domain.request.range.RequestRange;
import ch.psi.daq.query.bin.aggregate.BinnedValueCombiner;
import ch.psi.daq.queryrest.config.QueryRestConfig;
import ch.psi.daq.queryrest.model.HistoricChannelConfigurationPropertyFilterMixin;
import ch.psi.daq.queryrest.model.PropertyFilterMixin;
import ch.psi.daq.queryrest.response.AbstractHTTPResponse;
import ch.psi.daq.queryrest.response.ResponseFormatter;
import ch.psi.daq.queryrest.response.json.JSONResponseStreamWriter;

public class DAQQueriesResponseFormatter implements ResponseFormatter<List<Entry<DAQQueryElement, Stream<Triple<BackendQuery, ChannelName, ?>>>>>, ApplicationContextAware {
   private static final Logger LOGGER = LoggerFactory.getLogger(JSONResponseStreamWriter.class);

   public static final String DATA_RESP_FIELD = "data";

   public static final Mapping DEFAULT_MAPPING = new Mapping(IncompleteStrategy.PROVIDE_AS_IS);
   private static final long MILLIS_PER_PULSE = TimeUtils.MILLIS_PER_PULSE;
   private static final Function<DataEvent, ChannelName> KEY_PROVIDER = (event) -> new ChannelName(event.getChannel(),
         event.getBackend());
   // try to match sync data (bsread) with non sync data (epics) based on the time usin 10 millis
   // buckets.
   private static final ToLongFunction<DataEvent> MATCHER_PROVIDER = (event) -> event.getGlobalMillis()
         / MILLIS_PER_PULSE;
   
   // In case ArchiverAppliance had several events within the 10ms mapping interval, return these
   // aggregations (only used for table format)
   private Set<String> defaultEventResponseAggregations;
   
   @SuppressWarnings("unchecked")
   @Override
   public void setApplicationContext(ApplicationContext context) throws BeansException {
      final Set<Aggregation> defaultEventResponseAggregations =
            context.getBean(QueryRestConfig.BEAN_NAME_DEFAULT_EVENT_RESPONSE_AGGREGATIONS, Set.class);;
      this.defaultEventResponseAggregations =
            defaultEventResponseAggregations.stream().map(Aggregation::name)
                  .collect(Collectors.toCollection(LinkedHashSet::new));
   }

   @Override
   public void format(
         final JsonFactory factory,
         final ObjectMapper mapper,
         final List<Entry<DAQQueryElement, Stream<Triple<BackendQuery, ChannelName, ?>>>> results,
         final OutputStream out,
         final AbstractHTTPResponse response) throws Exception {
      final AtomicReference<Exception> exception = new AtomicReference<>();
      final JsonGenerator generator = factory.createGenerator(out, JsonEncoding.UTF8);

      try {
         if (results.size() > 1) {
            generator.writeStartArray();
         }

         results
               .forEach(entryy -> {
                  final DAQQueryElement daqQuery = entryy.getKey();

                  if (response.useTableFormat(daqQuery)) {
                     final Set<String> includedFields = getFields(daqQuery, false);
                     /* make sure identifiers are available */
                     includedFields.add(EventField.channel.name());
                     includedFields.add(EventField.backend.name());
                     // issue ATEST-633
                     if (!containsAggregation(includedFields)) {
                        includedFields.addAll(defaultEventResponseAggregations);
                     }

                     final ObjectWriter writer = configureWriter(mapper, includedFields, null);

                     writeTableFormat(generator, writer, entryy, exception);
                  } else {
                     final Set<String> includedFields = getFields(daqQuery, true);
                     final ObjectWriter writer = configureWriter(mapper, includedFields, null);

                     writeArrayFormat(generator, writer, entryy, exception);
                  }
               });
      } finally {
         if (results.size() > 1) {
            generator.writeEndArray();
         }

         generator.flush();
         generator.close();
      }

      if (exception.get() != null) {
         throw exception.get();
      }
   }

   private static void writeArrayFormat(final JsonGenerator generator, final ObjectWriter writer,
         final Entry<DAQQueryElement, Stream<Triple<BackendQuery, ChannelName, ?>>> entryy,
         final AtomicReference<Exception> exception) {
      final DAQQueryElement daqQuery = entryy.getKey();

      try {
         generator.writeStartArray();

         entryy.getValue()
               /* ensure elements are sequentially written */
               .sequential()
               .forEach(
                     triple -> {
                        try {
                           generator.writeStartObject();
                           generator.writeFieldName(EventField.channel.name());
                           writer.writeValue(generator, triple.getMiddle());
                           generator.writeFieldName(DATA_RESP_FIELD);
                           writer.writeValue(generator, triple.getRight());
                           generator.writeEndObject();
                        } catch (Exception e) {
                           LOGGER.error("Could not write channel name of channel '{}'", triple.getMiddle(),
                                 e);
                           exception.compareAndSet(null, e);
                        } finally {
                           if (triple.getRight() instanceof Stream) {
                              ((Stream<?>) (triple.getRight())).close();
                           }
                        }
                     });

         generator.writeEndArray();
      } catch (Exception e) {
         LOGGER.error("Exception while writing json for '{}'", daqQuery.getChannels(), e);
         exception.compareAndSet(null, e);
      }
   }

   @SuppressWarnings("unchecked")
   private static void writeTableFormat(JsonGenerator generator, ObjectWriter writer,
         Entry<DAQQueryElement, Stream<Triple<BackendQuery, ChannelName, ?>>> entryy,
         AtomicReference<Exception> exception) {
      /* get DataEvent stream of sub-queries for later match */
      final Map<ChannelName, Stream<DataEvent>> streams =
            new LinkedHashMap<>();
      final AtomicReference<BackendQuery> backendQueryRef = new AtomicReference<>();
      final DAQQueryElement daqQuery = entryy.getKey();

      entryy.getValue()
            .sequential()
            .forEach(
                  triple -> {
                     backendQueryRef.compareAndSet(null, triple.getLeft());

                     if (triple.getRight() instanceof Stream) {
                        streams.put(triple.getMiddle(), ((Stream<DataEvent>) triple.getRight()));
                     } else {
                        final String message =
                              String.format("Expect a DataEvent Stream for '%s' but got '%s'.",
                                    triple.getMiddle(), triple.getRight().getClass().getSimpleName());
                        LOGGER.warn(message);
                        streams.put(triple.getMiddle(), Stream.empty());
                     }
                  });

      final BackendQuery backendQuery = backendQueryRef.get();
      final RequestRange requestRange = backendQuery.getRequest().getRequestRange();
      BinningStrategy binningStrategy = backendQuery.getBinningStrategy();

      final Mapping mapping = daqQuery.getMappingOrDefault(DEFAULT_MAPPING);
      final Padder<ChannelName, DataEvent> padder = mapping.getIncomplete().getPadder(backendQuery);

      ToLongFunction<DataEvent> matchProvider = binningStrategy;
      if (binningStrategy == null) {
         matchProvider = MATCHER_PROVIDER;
         if (requestRange.isPulseIdRangeDefined()) {
            binningStrategy = new BinningStrategyPerBinPulse(1);
         } else if (requestRange.isTimeRangeDefined()) {
            binningStrategy = new BinningStrategyPerBinTime(MILLIS_PER_PULSE);
         } else {
            final String message = "Either time or pulseId range must be defined by the query!";
            LOGGER.error(message);
            throw new IllegalStateException(message);
         }
      }
      binningStrategy.setRequestRange(requestRange);

      /* online matching of the stream's content */
      final StreamMatcher<ChannelName, DataEvent, List<DataEvent>> streamMatcher =
            new StreamMatcher<>(
                  KEY_PROVIDER,
                  matchProvider,
                  new ListCreator<ChannelName, DataEvent>(),
                  new ListFiller<ChannelName, DataEvent>(),
                  new BinnedValueCombiner(binningStrategy),
                  padder,
                  streams.values());
      final Iterator<List<DataEvent>> streamsMatchIter = streamMatcher.iterator();

      try {
         generator.writeStartObject();
         generator.writeFieldName(DATA_RESP_FIELD);
         writer.writeValue(generator, streamsMatchIter);
         generator.writeEndObject();
      } catch (Exception e) {
         LOGGER.error("Exception while writing json for '{}'", daqQuery.getChannels(), e);
         exception.compareAndSet(null, e);
      } finally {
         if (streamMatcher != null) {
            try {
               streamMatcher.close();
            } catch (Throwable t) {
               LOGGER.error(
                     "Something went wrong while closing stream matcher for JSON table response writer.",
                     t);
            }
         }
      }
   }

   /**
    * Configures the writer dynamically by including the fields which should be included in the
    * response.
    * 
    * @param mapper The ObjectMapper
    * @param includedEventFields set of strings which correspond to the getter method names of the
    *        classes registered as a mixed-in
    * @param includedConfigFields set of strings which correspond to the getter method names of the
    *        classes registered as a mixed-in
    * @return the configured writer that includes the specified fields
    */
   public static ObjectWriter configureWriter(final ObjectMapper mapper, final Set<String> includedEventFields,
         final Set<String> includedConfigFields) {
      final SimpleFilterProvider propertyFilter = new SimpleFilterProvider();
      if (includedEventFields != null) {
         propertyFilter.addFilter(PropertyFilterMixin.FILTER_NAME,
               SimpleBeanPropertyFilter.filterOutAllExcept(includedEventFields));
      }
      if (includedConfigFields != null) {
         propertyFilter.addFilter(HistoricChannelConfigurationPropertyFilterMixin.FILTER_NAME,
               SimpleBeanPropertyFilter.filterOutAllExcept(includedConfigFields));
      }
      // only write the properties not excluded in the filter
      final ObjectWriter writer = mapper.writer(propertyFilter);
      return writer;
   }
   
   private static Set<String> getFields(final DAQQueryElement query, final boolean removeIdentifiers) {
      final Set<? extends QueryField> queryFields = query.getFields();
      final List<Aggregation> aggregations =
            query.getAggregation() != null ? query.getAggregation().getAggregations() : null;
      final List<Extrema> extrema = query.getAggregation() != null ? query.getAggregation().getExtrema() : null;

      final Set<String> includedFields =
            new LinkedHashSet<String>(queryFields.size() + (aggregations != null ? aggregations.size() : 0)
                  + (extrema != null ? extrema.size() : 0));

      for (final QueryField field : queryFields) {
         includedFields.add(field.getName());
      }
      if (aggregations != null) {
         for (final Aggregation aggregation : aggregations) {
            includedFields.add(aggregation.name());
         }
      }
      if (extrema != null) {
         // field of ExtremaCalculator (extrema in BinnedValueCombinedDataEvent and
         // BinnedIndexCombinedDataEvent)
         includedFields.add("extrema");
      }

      if (removeIdentifiers) {
         // do not write channel/backend since it is already provided as key in mapping
         includedFields.remove(EventField.channel.name());
         includedFields.remove(EventField.backend.name());
      }

      return includedFields;
   }
   
   private static boolean containsAggregation(final Set<String> includedFields) {
      for (final Aggregation aggregation : Aggregation.values()) {
         if (includedFields.contains(aggregation.name())) {
            return true;
         }
      }
      return false;
   }
}
